package timestamps;

import ij.IJ;
import ij.ImagePlus;
import ij.Prefs;
import ij.WindowManager;
import java.awt.Component;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Image;
import java.awt.Insets;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.KeyEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.util.ArrayList;
import java.util.EnumSet;
import java.util.Objects;
import javax.swing.Box;
import javax.swing.ButtonGroup;
import javax.swing.DefaultListCellRenderer;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JMenu;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JRadioButtonMenuItem;
import javax.swing.JSpinner;
import javax.swing.SpinnerNumberModel;
import javax.swing.border.EmptyBorder;
import timestamps.shared.IJFrame;
import timestamps.shared.IJLogger;
import timestamps.shared.IJLogger.Mode;
import timestamps.shared.MacroOptions;

public class Dialog extends IJFrame implements ActionListener, ItemListener {

   private static final String PREFS_BREAKS = "timestamps.breaks";
   private static final String PREFS_BREAK_NUMBER = "timestamps.break.number";
   private static final String PREFS_REPORT = "timestamps.report";
   private static final String PREFS_REPORT_TYPE = "timestamps.report.type";
   private static final String PREFS_CHANNEL_NUMBER = "timestamps.frame.number";
   private static final String PREFS_SLICE_NUMBER = "timestamps.slice.number";
   private static final String PREFS_SHOW = "timestamps.show";
   private static final String PREFS_AUTOSAVE = "timestamps.autosave";
   private static final String PREFS_OVERLAY = "timestamps.overlay";
   private static final String PREFS_EXIT = "timestamps.exit";
   private static final String PREFS_LOGGING = "timestamps.logging";

   private final Main main;
   private final IJLogger logger;
   private final JCheckBoxMenuItem exitBox;
   private final JLabel channelLabel, sliceLabel;
   private final JSpinner breaksSpinner, channelSpinner, sliceSpinner;
   private final SpinnerNumberModel breaksModel, channelModel, sliceModel;
   private final JCheckBox breaksBox, reportBox, displayBox, autosaveBox, overlayBox;
   private final JComboBox<Report> reportCombo;
   private final ArrayList<JComponent> reportComponents;

   /**Constructor.
    * @param main the executing model
    * @param logger the logger*/
   Dialog(Main main, IJLogger logger) {
      super(Main.TITLE);
      this.main = main;
      this.logger = logger;

      //Menubar
      exitBox = new JCheckBoxMenuItem("Exit when done");
      exitBox.setSelected(Prefs.get(PREFS_EXIT, false));
      exitBox.setToolTipText("Quits this dialog on completion");
      getMenu("Options").add(exitBox);

      JMenu menu = new JMenu("Logging");
      getMenu("Options").add(menu);
      ButtonGroup group = new ButtonGroup();
      MacroOptions<Mode> modes = new MacroOptions<>(Mode.class);
      MacroOptions<Report> reports = new MacroOptions<>(Report.class);
      ActionListener listener = evt -> {
         Mode mode = modes.getForCommand(evt.getActionCommand());
         switch (mode != null ? mode : Mode.INTERACTIVE) {
            case INTERACTIVE: logger.setInteractiveLogging(); break;
            case SILENT: logger.setSilentLogging(); break;
            case SYSTEM: logger.setSystemLogging(); break;
         }
         Prefs.set(PREFS_LOGGING, mode != null ? mode.getCommand() : Mode.INTERACTIVE.getCommand());
      };
      Mode preferred = modes.getForCommand(Prefs.get(PREFS_LOGGING, Mode.INTERACTIVE.getCommand()));
      for (Mode m : EnumSet.of(Mode.INTERACTIVE, Mode.SILENT, Mode.SYSTEM)) {
         JRadioButtonMenuItem button = new JRadioButtonMenuItem(m.getLabel());
         button.setToolTipText(m.getDescription());
         button.setActionCommand(m.getCommand());
         button.addActionListener(listener);
         group.add(button);
         menu.add(button);
         if (m == preferred) {
            button.setSelected(true);
            listener.actionPerformed(new ActionEvent(button, 0, m.getCommand()));
         }
      }

      //Content pane
      JPanel cp = new JPanel();
      setContentPane(cp);
      cp.setBorder(new EmptyBorder(10, 10, 10, 10));
      cp.setLayout(new GridBagLayout());
      GridBagConstraints c = new GridBagConstraints();
      c.fill = GridBagConstraints.NONE;
      c.anchor = GridBagConstraints.LINE_START;
      c.insets = new Insets(0, 0, 5, 5);

      boolean doBreaks = Prefs.get(PREFS_BREAKS, true);
      breaksBox = createBox(Option.BREAKS.getLabel(), doBreaks, true,
            Option.BREAKS.getDescription());
      int breaks = Prefs.getInt(PREFS_BREAK_NUMBER, 1);
      breaksModel = new SpinnerNumberModel(breaks, 1, 10, 1);
      breaksSpinner = new JSpinner(breaksModel);
      breaksSpinner.setEnabled(doBreaks);
      c.gridx = 0;
      c.gridy = 0;
      cp.add(breaksBox, c);
      c.gridx ++;
      cp.add(breaksSpinner, c);

      boolean doReport = Prefs.get(PREFS_REPORT, true);
      reportBox = createBox(Option.REPORT.getLabel(), doReport, true,
            Option.REPORT.getDescription());
      c.gridx = 0;
      c.gridy ++;
      cp.add(reportBox, c);

      reportCombo = new JComboBox<>(reports.getAllAsArray());
      reportCombo.setSelectedItem(Prefs.get(PREFS_REPORT_TYPE, Report.FRAME.getLabel()));
      DefaultListCellRenderer renderer = new DefaultListCellRenderer() {

         @Override
         public Component getListCellRendererComponent(JList list, Object value,
               int index, boolean isSelected, boolean cellHasFocus) {
            Report report = (Report) value;
            JComponent c = (JComponent) super.getListCellRendererComponent(list,
                  value, index, isSelected, cellHasFocus);
            c.setToolTipText(report.getDescription());
            c.setEnabled(reportCombo.isEnabled());
            return c;
         }
      };
      reportCombo.setRenderer(renderer);
      reportCombo.setEnabled(doReport);
      reportCombo.addActionListener(this);
      reportComponents = new ArrayList<>(3);
      reportComponents.add(reportCombo);
      c.gridx ++;
      c.gridwidth = 2;
      cp.add(reportCombo, c);

      channelLabel = new JLabel("Channel");
      reportComponents.add(channelLabel);
      channelModel = new SpinnerNumberModel(Prefs.get(PREFS_CHANNEL_NUMBER, 1),
            1, 100, 1);
      channelSpinner = new JSpinner(channelModel);
      channelSpinner.setEnabled(true);
      reportComponents.add(channelSpinner);
      c.gridx = 1;
      c.gridy ++;
      cp.add(channelLabel, c);
      c.gridx ++;
      cp.add(channelSpinner, c);

      sliceLabel = new JLabel("Slice");
      reportComponents.add(sliceLabel);
      sliceModel = new SpinnerNumberModel(Prefs.get(PREFS_CHANNEL_NUMBER, 1),
            1, 300, 1);
      sliceSpinner = new JSpinner(sliceModel);
      sliceSpinner.setEnabled(true);
      reportComponents.add(sliceSpinner);
      c.gridx = 1;
      c.gridy ++;
      cp.add(sliceLabel, c);
      c.gridx ++;
      cp.add(sliceSpinner, c);

      displayBox = createBox(Option.DISPLAY.getLabel(), Prefs.get(PREFS_SHOW, true),
            false, Option.DISPLAY.getDescription());
      displayBox.setEnabled(doReport);
      reportComponents.add(displayBox);
      c.gridx = 0;
      c.gridy ++;
      cp.add(displayBox, c);

      autosaveBox = createBox(Option.AUTOSAVE.getLabel(), Prefs.get(PREFS_AUTOSAVE, false),
            false,Option.AUTOSAVE.getDescription());
      autosaveBox.setEnabled(doReport);
      reportComponents.add(autosaveBox);
      c.gridx = 0;
      c.gridy ++;
      cp.add(autosaveBox, c);

      overlayBox = createBox(Option.OVERLAY.getLabel(), Prefs.get(PREFS_OVERLAY, false),
            false, Option.OVERLAY.getDescription());
      reportComponents.add(overlayBox);
      c.gridx = 0;
      c.gridy ++;
      c.gridwidth = 2;
      cp.add(overlayBox, c);

      c.gridx = 0;
      c.gridy ++;
      cp.add(Box.createVerticalStrut(5), c);

      JButton readButton = new JButton("Read");
      readButton.addActionListener(this);
      readButton.setMnemonic(KeyEvent.VK_R);
      c.gridy ++;
      c.fill = GridBagConstraints.HORIZONTAL;
      c.gridwidth = 4;
      cp.add(readButton, c);

      pack();
      IJ.getInstance().addWindowListener(new WindowAdapter() {

         @Override
         public void windowClosing(WindowEvent evt) {
            quit();
         }
      });
   }

   @Override
   public void actionPerformed(ActionEvent evt) {
      if (evt.getSource().equals(reportCombo)) {
         String selected = Objects.requireNonNull(reportCombo.getSelectedItem()).toString();
         if (selected.equals(Report.FRAME.getLabel())) {
            channelLabel.setEnabled(true);
            channelSpinner.setEnabled(true);
            sliceLabel.setEnabled(true);
            sliceSpinner.setEnabled(true);
         } else if (selected.equals(Report.CHANNEL.getLabel())) {
            channelLabel.setEnabled(false);
            channelSpinner.setEnabled(false);
            sliceLabel.setEnabled(true);
            sliceSpinner.setEnabled(true);
         } else if (selected.equals(Report.SLICE.getLabel())) {
            channelLabel.setEnabled(true);
            channelSpinner.setEnabled(true);
            sliceLabel.setEnabled(false);
            sliceSpinner.setEnabled(false);
         } else if (selected.equals(Report.ALL.getLabel())) {
            channelLabel.setEnabled(false);
            channelSpinner.setEnabled(false);
            sliceLabel.setEnabled(false);
            sliceSpinner.setEnabled(false);
         }
      } else {
         //Read button
         ImagePlus imp = WindowManager.getCurrentImage();
         if (imp == null) {
            JOptionPane.showMessageDialog(null, "There is no image open",
                  Main.TITLE, JOptionPane.ERROR_MESSAGE, null);
            return;
         }
         //Generate the command
         StringBuilder sb = new StringBuilder();
         if (breaksBox.isSelected()) {
            sb.append(Option.BREAKS.getCommand());
            sb.append("=");
            sb.append(breaksModel.getNumber().intValue());
            sb.append(" ");
         }
         if (reportBox.isSelected()) {
            sb.append(Option.REPORT.getCommand());
            sb.append("=");
            Report selection = (Report) reportCombo.getSelectedItem();
            if (selection == null) {
               sb.append(Report.FRAME.getCommand()); //default
            } else {
               sb.append(selection.getCommand());
            }
            sb.append(" ");
            if (selection == Report.FRAME || selection == Report.SLICE) {
               sb.append(Option.CHANNEL.getCommand());
               sb.append("=");
               sb.append(channelModel.getNumber().intValue());
               sb.append(" ");
            }
            if (selection == Report.FRAME || selection == Report.CHANNEL) {
               sb.append(Option.SLICE.getCommand());
               sb.append("=");
               sb.append(sliceModel.getNumber().intValue());
               sb.append(" ");
            }
            if (displayBox.isSelected()) {
               sb.append(Option.DISPLAY.getCommand());
               sb.append(" ");
            }
            if (autosaveBox.isSelected()) {
               sb.append(Option.AUTOSAVE.getCommand());
               sb.append(" ");
            }
            if (overlayBox.isSelected()) {
               sb.append(Option.OVERLAY.getCommand());
            }
         }
         main.process(sb.toString().trim());

         //Store the options in the preferences
         Prefs.set(PREFS_BREAKS, breaksBox.isSelected());
         Prefs.set(PREFS_BREAK_NUMBER, (Integer) breaksModel.getValue());
         Prefs.set(PREFS_REPORT, reportBox.isSelected());
         Report report = (Report) reportCombo.getSelectedItem();
         if (report != null) {
            Prefs.set(PREFS_REPORT_TYPE, report.getCommand());
         }
         Prefs.set(PREFS_CHANNEL_NUMBER, channelModel.getNumber().intValue());
         Prefs.set(PREFS_SLICE_NUMBER, sliceModel.getNumber().intValue());
         Prefs.set(PREFS_SHOW, displayBox.isSelected());
         Prefs.set(PREFS_AUTOSAVE, autosaveBox.isSelected());
         Prefs.set(PREFS_OVERLAY, overlayBox.isSelected());
         Prefs.set(PREFS_EXIT, exitBox.isSelected());
         Prefs.savePreferences();
         if (exitBox.isSelected()) {
            quit();
         }
      }
   }

   @Override
   public void itemStateChanged(ItemEvent evt) {
      Object source = evt.getSource();
      boolean selected = false;
      if (evt.getStateChange() == ItemEvent.SELECTED) {
         selected = true;
      }
      if (source == reportBox) {
         for (JComponent jc : reportComponents) {
            jc.setEnabled(selected);
         }
      } else if (source == breaksBox) {
         breaksSpinner.setEnabled(selected);
      }
   }

   @Override
   protected void displayAbout() {
      Image icon = Toolkit.getDefaultToolkit().getImage(Dialog.class.getResource("/biooptics.png"));
      JOptionPane.showMessageDialog(null,
                  "\u00A9 2020, Thomas Peterbauer, University of Vienna\n\n"
                        +"Version "+Main.VERSION+"\n\n"
                  +"Contains the 'Strings' comparator library written\n by Stephen Kelvin Friedrich (2006)",
            Main.TITLE, JOptionPane.INFORMATION_MESSAGE, new ImageIcon(icon));
   }

   @Override
   protected void quit() {
      logger.cleanup();
      super.quit();
   }

   /**Creates a new Checkbox
    * @param label the label to appear on the UI
    * @param setSelected set selected or not
    * @param addListener add an {@link ItemListener} or not
    * @param tooltip the tooltip*/
   private JCheckBox createBox(String label, boolean setSelected, boolean addListener, String tooltip) {
      JCheckBox box = new JCheckBox(label);
      box.setSelected(setSelected);
      if (addListener) {
         box.addItemListener(this);
      }
      box.setToolTipText(tooltip);
      return box;
   }
}
