package timestamps;

import ij.IJ;
import ij.ImagePlus;
import ij.Macro;
import ij.plugin.PlugIn;
import ij.plugin.frame.Recorder;
import loci.common.DebugTools;
import timestamps.Timestamps.TimestampImpl;
import timestamps.parser.Deltavision;
import timestamps.parser.OMEXML;
import timestamps.parser.Parser;
import timestamps.parser.ParserException;
import timestamps.parser.VisiView;
import timestamps.parser.cellSens;
import timestamps.shared.IJLogger;
import timestamps.shared.MacroOptions;
import timestamps.shared.Metadata;

/**Main class of the ImageJ timestamp reader*/
public class Main implements PlugIn {

   static final String TITLE = "Timestamp Reader";
   static final String VERSION = "1.0.2";
   private IJLogger logger;

   /**Called on loading. The plugin is macro scriptable. Arguments can be passed by running the command:<br>
    * <br>
    *       <code>run("Timestamps", "arg1 arg2...");</code><br>
    * <br>
    * For the syntax, use the ImageJ macro recorder.
    * @param s unused (no parameters from plugins.config implemented)*/
   @Override
   public void run(String s) {
      //Sample image for testing
//      try {
//         ImporterOptions options = new ImporterOptions();
//         options.setId("some_path");
//         options.setSeriesOn(0, true);
//         ImagePlus[] imps = BF.openImagePlus(options);
//         imps[0].show();
//      } catch (FormatException e) {
//         e.printStackTrace();
//      } catch (IOException e) {
//         e.printStackTrace();
//      }

      logger = new IJLogger(TITLE);
      //Assert that Bio-Formats is present
      try {
         Class.forName("loci.plugins.BF");
         DebugTools.setRootLevel("ERROR");
      } catch (ClassNotFoundException e) {
         logger.error("Required Bio-Formats plugin not found");
      }
      //Show a dialog if launched without arguments (UI mode), or process the arguments
      String args = Macro.getOptions();
      if (args == null || args.isEmpty()) {
         new Dialog(this, logger).setVisible(true);
      } else {
         process(args);
      }
   }

   /**Processes an image (scriptable and recordable)
    * @param args IJ macro arguments*/
   void process(String args) {
      //Argument parsing
      if (args == null || args.isEmpty()) {
         logger.warning("No macro arguments provided");
         return;
      }
      MacroOptions<Option> options = new MacroOptions<>(Option.class, args, logger);
      //Get the active image and run some checks
      ImagePlus imp = IJ.getImage();
      if (imp == null) {
         return; //notice the user here?
      }
      if (imp.getNFrames() < 2) {
         logger.error(imp.getTitle()+" is not a time series");
         return;
      }
      Metadata metadata = new Metadata(imp);
      if (metadata.getOriginalFileInfo() == null) {
         logger.error(imp.getTitle()+" has no original metadata");
         return;
      }
      if (metadata.isEmpty()) {
         logger.error(imp.getTitle()+" has not enough metadata");
         return;
      }
      //Search for the appropriate parser
      Parser parser = null;
      if (metadata.getString("Software").contains("VisiView")) {
         parser = new VisiView();
      } else if (metadata.getString("Product Name #1").contains("OLYMPUS cellSens")) {
         parser = new cellSens();
      } else if (metadata.getOriginalFileInfo().fileName.endsWith(".dv")) {
         parser = new Deltavision();
      } else {
         String app = metadata.getString("Information|Application|Name #1");
         if (app.contains("ZEN") && app.contains("blue edition")) {
            parser = new OMEXML();
         }
      }
      TimestampImpl timestamps = null;
      if (parser != null) {
         try {
            timestamps = Timestamps.create(metadata, logger);
            parser.parse(metadata, timestamps);
         } catch (ParserException e) {
            logger.error(e);
         }
      } else {
         logger.error("No suitable parser found for "+imp.getTitle());
         return;
      }
      timestamps.compute(options);
      //Macro recorder
      if (Recorder.record) {
         StringBuilder sb = new StringBuilder("run(\"Timestamps\", \"");
         sb.append(options.toCommandLine());
         sb.append("\");");
         //Append a newline character (at least on Linux the Recorder does not do this on its own)
         sb.append(System.lineSeparator());
         Recorder.recordString(sb.toString());
      }
   }
}
