package timestamps;

import java.time.Duration;
import java.time.temporal.ChronoUnit;

/**Wrapper around {@link ChronoUnit} to format {@link Duration} objects, up to {@link ChronoUnit#HOURS} and
 * down to {@link ChronoUnit#MILLIS}<br><br>
 * <i>Note:</i> this class should be refactored once Fiji supports JDK9+. Later versions of {@link Duration}
 * contain improved methods for formatting*/
enum Time {

   HOURS("h", ChronoUnit.HOURS) {

      /**{@inheritDoc}*/
      @Override
      void format(StringBuilder sb, Duration d, int padding, char charachter, Time shortest) {
         long hours = d.toHours();
         append(sb, hours, padding, charachter);
         sb.append(DELIMITER);
         MINUTES.format(sb, d.minus(hours, ChronoUnit.HOURS), 2, '0', shortest);
      }

      /**{@inheritDoc}*/
      @Override
      Time getNextSmaller() {
         return Time.MINUTES;
      }

      /**{@inheritDoc}
       * @return the duration as a fraction of hours*/
      @Override
      float toFloat(Duration d) {
         return d.getSeconds() / 3600f;
      }
   },

   MINUTES("min", ChronoUnit.MINUTES) {

      /**{@inheritDoc}*/
      @Override
      void format(StringBuilder sb, Duration d, int padding, char character, Time shortest) {
         long minutes = d.toMinutes();
         if (shortest == Time.MINUTES) {
            //Round the residual seconds to the next minute
            long seconds = d.minus(minutes, ChronoUnit.MINUTES).getSeconds();
            if (seconds > 30) {
               minutes += 1;
            }
         }
         append(sb, minutes, padding, character);
         if (shortest != Time.MINUTES) {
            sb.append(DELIMITER);
            SECONDS.format(sb, d.minus(minutes, ChronoUnit.MINUTES), 2, '0', shortest);
         }
      }

      /**{@inheritDoc}*/
      @Override
      Time getNextSmaller() {
         return Time.SECONDS;
      }

      /**{@inheritDoc}*/
      @Override
      float toFloat(Duration d) {
         return d.toMillis() / 60000f;
      }
   },

   SECONDS("s", ChronoUnit.SECONDS) {

      /**{@inheritDoc}*/
      @Override
      void format(StringBuilder sb, Duration d, int padding, char character, Time shortest) {
         long seconds = d.getSeconds();
         if (shortest == Time.SECONDS) {
            //Round the residual milliseconds to the next second
            long millis = d.minus(seconds, ChronoUnit.SECONDS).toMillis();
            if (millis > 500) {
               seconds += 1;
            }
         }
         append(sb, seconds, padding, character);
         if (shortest != Time.SECONDS) {
            sb.append(".");
            MILLISECONDS.format(sb, d.minus(seconds, ChronoUnit.SECONDS), 3, '0', shortest);
         }
      }

      /**{@inheritDoc}*/
      @Override
      Time getNextSmaller() {
         return Time.MILLISECONDS;
      }

      /**{@inheritDoc}*/
      @Override
      float toFloat(Duration d) {
         return (float) d.getSeconds() + d.getNano() / 1000000000f;
      }
   },

   MILLISECONDS("ms", ChronoUnit.MILLIS) {

      /**{@inheritDoc}*/
      @Override
      void format(StringBuilder sb, Duration d, int padding, char character, Time shortest) {
         long millis = d.toMillis();
         append(sb, millis, padding, character);
      }

      /**{@inheritDoc}*/
      @Override
      Time getNextSmaller() {
         return null;
      }

      /**{@inheritDoc}*/
      @Override
      float toFloat(Duration d) {
         return (float) d.getSeconds() * 1000f + d.getNano() / 1000000f;
      }
   };

   private static final String DELIMITER = ":";

   final String SYMBOL;
   private final ChronoUnit TIME_UNIT;

   /**Constructor
    * @param symbol short lower case notation of the time unit
    * @param timeUnit the time unit*/
   Time(String symbol, ChronoUnit timeUnit) {
      this.SYMBOL = symbol;
      this.TIME_UNIT = timeUnit;
   }

   /**Appends a number to a string builder, padded with the supplied character(s)
    * @param sb the string builder
    * @param number the number
    * @param padding the total number of digits
    * @param character the character for padding*/
   protected void append(StringBuilder sb, long number, int padding, char character) {
      String str = String.valueOf(number);
      if (str.length() == padding) {
         sb.append(str);
      } else {
         for (int i = 0; i < (padding - str.length()); i++) {
            sb.append(character);
         }
         sb.append(str);
      }
   }

   /**{@inheritDoc}*/
   @Override
   public String toString() {
      return TIME_UNIT.toString();
   }

   /**Computes the largest non-null time unit of a duration up to hours (i.e., no days/months etc.) and
    * down to milliseconds
    * @param d the duration
    * @return the time unit*/
   @SuppressWarnings("unused")
   static Time getDimension(Duration d) {
      if (d.toHours() > 0) {
         return Time.HOURS;
      } else if (d.toMinutes() > 0) {
         return Time.MINUTES;
      } else if (d.getSeconds() > 0) {
         return Time.SECONDS;
      } else {
         return Time.MILLISECONDS;
      }
   }

   /**Formats a duration into human-readable form
    * @param d the duration
    * @param padding of the largest time unit with zeros (keeps timestamps in images on a constant position)
    * @param character the character for padding (usually zero or a white space)
    * @param shortest the smallest time unit to be displayed (e.g., {@link Time#MINUTES} for "HH:mm"
    *    rather than HH:mm:ss.SSS""
    * @return the formatted string or <code>null</code>> if the duration provided is <code>null</code>*/
   @SuppressWarnings("unused")
   static String format(Duration d, int padding, char character, Time longest, Time shortest) {
      if (d == null) {
         return null;
      }
      StringBuilder sb = new StringBuilder();
      longest.format(sb, d, padding, character, shortest);
      return sb.toString();
   }

   /**Returns the fractional value of a duration in the smallest time unit (yielding a value > 1). The unit
    * can be computed with {@link Time#getDimension(Duration)}.
    * @param d the duration
    * @return formatted value*/
   @SuppressWarnings("unused")
   static float toSmallestFloat(Duration d) {
      if (d.toHours() > 0) {
         return Time.HOURS.toFloat(d);
      } else if (d.toMinutes() > 0) {
         return Time.MINUTES.toFloat(d);
      } else if (d.getSeconds() > 0) {
         return Time.SECONDS.toFloat(d);
      } else {
         return MILLISECONDS.toFloat(d);
      }
   }

   /**Appends a formatted duration to the string builder provided<br>
    * <br>
    * <i>Note: </i> this method is for internal use and package-protected only for implementation reasons.
    * Use {@link Time#format(Duration, int, char, Time, Time)}
    * @param sb the string builder
    * @param d the duration
    * @param padding minimum digits of the largest time (keeps timestamps in images on a constant position)
    * @param character the character for padding (usually zero or a white space)
    * @param shortest the smallest time unit to be displayed (e.g., {@link Time#MINUTES} for "HH:mm"
    *    rather than HH:mm:ss.SSS""*/
   @SuppressWarnings("unused")
   abstract void format(StringBuilder sb, Duration d, int padding, char character, Time shortest);

   /**Returns the next smaller time unit (minutes for hours, seconds for minutes and so on)
    * @return the next smaller unit, or <code>null</code> if there is none implemented*/
   @SuppressWarnings("unused")
   abstract Time getNextSmaller();

   /**Computes the fraction of a duration
    * @param d the duration
    * @return the fraction*/
   @SuppressWarnings("unused")
   abstract float toFloat(Duration d);
}
