package timestamps;

import ij.ImagePlus;
import ij.WindowManager;
import ij.gui.Overlay;
import ij.gui.Roi;
import ij.gui.TextRoi;
import ij.io.FileInfo;
import ij.measure.Measurements;
import ij.measure.ResultsTable;
import ij.process.ImageStatistics;
import java.awt.Color;
import java.awt.Font;
import java.awt.Frame;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.time.Duration;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.ListIterator;
import timestamps.shared.IJLogger;
import timestamps.shared.MacroOptions;
import timestamps.shared.Metadata;

/**Interface for the timestamps to be generated by parsers.*/
public interface Timestamps {

   /**Creates a default implementation of the {@link Timestamps} interface for parsers
    * @param metadata the metadata of the image
    * @param logger the logger
    * @return the implementation*/
   static TimestampImpl create(Metadata metadata, IJLogger logger) {
      return new TimestampImpl(metadata, logger);
   }

   /**Sets channel names, if available.
    * @param index the index of a channel, starting at 0
    * @param name the name of the channel*/
   void setChannelName(int index, String name);

   /**Sets the series name. This should be the filename with a suffix for position and/or channel (without
    * extension). If parsers do not set it, the file name from the image's {@link FileInfo} will be used as
    * name for reports (which may result in errors or overwritten files if the series is not included).
    * @param series the name of the current series*/
   void setSeriesName(String series);

   /**Parsers must use this method to set timestamps of individual images, reported as durations from the
    * sequence start to the acquisition.
    * @param channel the channel (0 for single-channel images)
    * @param slice the z slice (0 for single-slice images)
    * @param frame the time point, with index 0 being the first image (<i>not</i> the start of a sequence!)
    * @param time the duration*/
   void setDuration(int channel, int slice, int frame, Duration time);

   /**Implementation of a container for timestamps.*/
   class TimestampImpl implements Timestamps {

      /**Prefix for results table titles*/
      private static final String PREFIX = "Timestamps-";
      /**Default title for results table z slice columns*/
      private static final String Z = "Z-Slice";
      /**Key set as property of all {@link TextRoi} objects to identify timestamp overlays*/
      private static final String PROPERTY_KEY = "type";
      /**Property value to identify timestamp overlays*/
      private static final String PROPERTY_VALUE ="timestamp";

      private final Metadata metadata;
      private final IJLogger logger;
      private final String[] channelNames;
      private final int nChannels;
      private final int nSlices;
      private final int nFrames;
      /**Multidimensional array for the timestamps, in the order channel/slice/frame.*/
      private final Duration[][][] durations;

      private String seriesName;

      /**Private constructor.
       * @param metadata the metadata of the image
       * @param logger the logger*/
      private TimestampImpl(Metadata metadata, IJLogger logger) {
         this.metadata = metadata;
         this.logger = logger;
         channelNames = new String[metadata.getImagePlus().getNChannels()];
         nChannels = metadata.getImagePlus().getNChannels();
         nSlices = metadata.getImagePlus().getNSlices();
         nFrames = metadata.getImagePlus().getNFrames();
         durations = new Duration[nChannels][nSlices][nFrames];
      }

      /**{@inheritDoc}*/
      @Override
      public void setChannelName(int index, String name) {
         channelNames[index] = name;
      }

      /**{@inheritDoc}*/
      @Override
      public void setSeriesName(String series) {
         this.seriesName = series;
      }

      /**{@inheritDoc}*/
      @Override
      public void setDuration(int channel, int slice, int frame, Duration time) {
         durations[channel][slice][frame] = time;
      }

      /**Processes data generated by parsers.
       * @param options the parsed macro arguments*/
      @SuppressWarnings("ConstantConditions")
      void compute(MacroOptions<Option> options) {
         //Set dfault channel names if not provided by the parser
         for (int i = 1; i <= channelNames.length; i++) {
            if (channelNames[i - 1] == null) {
               channelNames[i - 1] = "C"+i;
            }
         }

         //Set a default channel and slice and define a duration to be set zero (default: the first image)
         int useChannel = 0;
         int useSlice = 0;
         Duration zero = durations[0][0][0];

         Report report = (Report) options.getParameter(Option.REPORT);
         if (report != null) {
            //Downgrade the report option if there are is only one channel or slice
            if ((report == Report.CHANNEL && nChannels == 1) ||
                  (report == Report.SLICE && nSlices == 1)) {
               report = Report.FRAME;
            } else if (report == Report.ALL) {
               if (nChannels == 1 && nSlices == 1) {
                  report = Report.FRAME;
               } else if (nChannels == 1) {
                  report = Report.SLICE;
               } else if (nSlices == 1) {
                  report = Report.CHANNEL;
               }
            }
            if ((report == Report.SLICE || report == Report.FRAME) && options.contains(Option.CHANNEL)) {
               useChannel = options.getInteger(Option.CHANNEL);
               useChannel -= 1; //in the UI, channels start at 1, we here at 0
               if (nChannels <= useChannel) {
                  useChannel = nChannels - 1;
               }
               useChannel = Math.max(useChannel, 0); //just in case
            }
            if ((report == Report.CHANNEL || report == Report.FRAME) && options.contains(Option.SLICE)) {
               useSlice = options.getInteger(Option.SLICE);
               if (nSlices <= useSlice) {
                  useSlice = nSlices - 1;
               }
               useSlice -= 1; //in the UI, slices start at 1, we here at 0
               useSlice = Math.max(useSlice, 0);
            }
            zero = durations[useChannel][useSlice][0];
         }

         //Compute intervals from the first time point (i.e., set the first frame to zero)
         if (report == Report.ALL) {
            //Determine the smallest duration of all images of the first frame
            for (int i = 0; i < nChannels; i++) {
               for (int j = 0; j < nSlices; j++) {
                  if (durations[i][j][0].compareTo(zero) < 0) {
                     zero = durations[i][j][0];
                  }
               }
            }
         } else if (report == Report.CHANNEL) {
            //Determine the smallest duration of all channels of the first frame, at the slice to use
            for (int i = 0; i < nChannels; i++) {
               if (durations[i][useSlice][0].compareTo(zero) < 0) {
                  zero = durations[i][useSlice][0];
               }
            }
         } else if (report == Report.SLICE) {
            //Determine the smallest duration of all slices of the first frame, at the channel to use
            for (int i = 0; i < nSlices; i++) {
               if (durations[useChannel][i][0].compareTo(zero) < 0) {
                  zero = durations[useChannel][i][0];
               }
            }
         }
         Duration[][][] d = new Duration[nChannels][nSlices][nFrames];
         for (int i = 0; i < nChannels; i++) {
            for (int j = 0; j < nSlices; j++) {
               for (int k = 0; k < nFrames; k++) {
                  d[i][j][k] = durations[i][j][k].minus(zero);
               }
            }
         }

         //Calculate intervals between frames (always needed - we need to know the smallest time unit)
         List<Duration> intervals = new ArrayList<>();
         for (int i = 0; i < (nFrames - 1); i++) {
            intervals.add(d[useChannel][useSlice][i + 1].minus(d[useChannel][useSlice][i]));
         }
         /* Simple outlier test based on interquartile ranges. See http://www.mathwords.com/o/outlier.htm
          * Implementation as in https://stackoverflow.com/questions/18805178/
          * Skip if there are < 4 intervals or if the break number is unrealistically high (>25% of the
          * intervals)*/
         if (options.contains(Option.BREAKS)) {
            int breaks = options.getInteger(Option.BREAKS);
            if (intervals.size() <= 3) {
               logger.info("Skipping outlier test for breaks (less than 4 intervals available)");
            } else if (breaks / (float) intervals.size() > 0.25f) {
               logger.info("Skipping outlier test (break number higher than 25% of all time points");
            } else {
               Collections.sort(intervals);
               List<Duration> intervals1;
               List<Duration> intervals2;
               if (intervals.size() % 2 == 0) {
                  intervals1 = intervals.subList(0, intervals.size() / 2);
                  intervals2 = intervals.subList(intervals.size() / 2, intervals.size());
               } else {
                  intervals1 = intervals.subList(0, intervals.size() / 2);
                  intervals2 = intervals.subList(intervals.size() / 2 + 1, intervals.size());
               }
               Duration q1 = getMedian(intervals1);
               Duration q3 = getMedian(intervals2);
               if (intervals2.size() == 2) {
                  /* Even-numbered list with only 2 elements. Do not average 2 values around the center
                  as done
                   * in the getMedian(List) method, but use the lower value. With averaging, even a very
                    large
                   * outlier is never detected*/
                  q3 = intervals2.get(0);
               }
               Duration iqr = q3.minus(q1);
               /* Duration lowerFence = q1.minus(iqr.multipliedBy((long) 1.5));
                * Ignore the lower fence (breaks will always increase the time interval)*/
               Duration upperFence = q3.plus(iqr.multipliedBy((long) 1.5));
               int count = 0;
               ListIterator<Duration> it = intervals.listIterator(intervals.size());
               while (it.hasPrevious() && count <= breaks) {
                  Duration duration = it.previous();
                  if (duration.compareTo(upperFence) > 0) {
                     it.remove();
                     count++;
                  }
               }
            }
         }

         //Compute the average interval
         Duration sum = Duration.ZERO;
         for (Duration duration : intervals) {
            sum = sum.plus(duration);
         }
         Duration interval = sum.dividedBy(intervals.size());
         Time unit = Time.getDimension(interval);

         //Set the image calibration (see menu command Image -> Properties)
         metadata.getImagePlus().getCalibration().setTimeUnit(unit.SYMBOL);
         metadata.getImagePlus().getCalibration().frameInterval = Time.toSmallestFloat(interval);

         //Exit if there are not report tasks required
         if (report == null) {
            return;
         }

         //Create the table
         if (options.contains(Option.DISPLAY) || options.contains(Option.AUTOSAVE)) {
            //Delete a previous results table with the same title
            String title = PREFIX+seriesName;
            for (String s : WindowManager.getNonImageTitles()) {
               if (s.equals(title)) {
                  Frame f = WindowManager.getFrame(title);
                  f.dispose();
                  WindowManager.removeWindow(f);
               }
            }

            ResultsTable table = new ResultsTable();
            table.setValue("Frame", 0, 1); //frames are always in the first column
            switch (report) {
               case FRAME:
                  if (nSlices > 1) {
                     table.setValue(Z, 0, useSlice+1);
                  } else {
                     table.setValue(Z, 0, 1);
                  }
                  String header = "T_"+unit.SYMBOL;
                  if (nChannels > 1) {
                     header = channelNames[useChannel]+"_"+unit.SYMBOL;
                  }
                  table.setValue(header, 0, 0);
                  for (int i = 0; i < nFrames; i++) {
                     table.setValue(0, i, i+1); //frame number, starting at 1
                     if (nSlices > 1) {
                        table.setValue(1, i, useSlice+1);
                        table.setValue(2, i, unit.toFloat(d[useChannel][useSlice][i]));
                     } else {
                        table.setValue(1, i, 1);
                        table.setValue(2, i, unit.toFloat(d[useChannel][0][i]));
                     }
                  }
                  break;
               case CHANNEL:
                  if (nSlices > 1) {
                     table.setValue(Z, 0, useSlice+1);
                  } else {
                     table.setValue(Z, 0, 1);
                  }
                  for (int i = 0; i < nFrames; i++) {
                     table.setValue(0, i, i+1); //frame number
                     if (nSlices > 1) {
                        table.setValue(1, i, useSlice+1); //slice info
                     } else {
                        table.setValue(1, i, 1);
                     }
                  }
                  for (int i = 0; i < nChannels; i++) {
                     table.setValue(channelNames[i]+"_"+unit.SYMBOL, 0, 0);
                     for (int j = 0; j < nFrames; j++) {
                        if (nSlices > 1) {
                           table.setValue(i+2, j, unit.toFloat(d[i][useSlice][j])); //time
                        } else {
                           table.setValue(i+2, j, unit.toFloat(d[i][0][j]));
                        }
                     }
                  }
                  break;
               case SLICE:
                  for (int i = 0; i < nFrames; i++) {
                     table.setValue(0, i, i+1); //frame number
                  }
                  if (nChannels > 1) { //channel (skip if there's only one)
                     table.setValue("C", 0, channelNames[useChannel]);
                     for (int i = 0; i < nFrames; i++) {
                        table.setValue(1, i, channelNames[useChannel]);
                     }
                  }
                  for (int i = 0; i < nSlices; i++) {
                     int slice = i + 1;
                     table.setValue("Z"+slice+"_"+unit.SYMBOL, 0, 0);
                     for (int j = 0; j < nFrames; j++) {
                        if (nChannels > 1) {
                           table.setValue(i+2, j, unit.toFloat(d[useChannel][i][j]));
                        } else {
                           table.setValue(i+1, j, unit.toFloat(d[0][i][j]));
                        }
                     }
                  }
                  break;
               case ALL:
                  table.setValue(Z, 0, 1);
                  for (int i = 0; i < nChannels; i++) {
                     table.setValue(channelNames[i]+"_"+unit.SYMBOL, 0, 0);
                  }
                  for (int i = 0; i < nFrames; i++) {
                     for (int j = 0; j < nSlices; j++) {
                        int n = i * nSlices + j;
                        int frame = i + 1;
                        int slice = j + 1;
                        table.setValue(0, n, frame);
                        table.setValue(1, n, slice);
                        for (int k = 0; k < nChannels; k++) {
                           table.setValue(k+2, n, unit.toFloat(d[k][j][i]));
                        }
                     }
                  }
                  break;
            }
            //Display the table
            if (options.contains(Option.DISPLAY)) {
               table.show(PREFIX+seriesName);
            }
            //Autosaving without overwriting of existing files
            if (options.contains(Option.AUTOSAVE)) {
               if (seriesName == null || seriesName.isEmpty()) {
                  seriesName = metadata.getOriginalFileInfo().fileName;
               }
               String suffix = "_timestamps";
               Path dir = Paths.get(metadata.getOriginalFileInfo().directory);
               Path path = dir.resolve(seriesName+suffix+".txt");
               try {
                  if (Files.exists(path)) {
                     int count = 0;
                     do {
                        if (count >= 5) {
                           logger.warning("Gave up writing to "+path.toString()+" (5 attempts)");
                           break;
                        }
                        count ++;
                        path = dir.resolve(seriesName+suffix+"_"+count+".txt");
                     } while (Files.exists(path));
                  }
                  table.saveAs(path.toString());
                  logger.info("Timestamps saved as "+path.getFileName().toString());
               } catch (IOException e) {
                  logger.error(e);
               }
            }
         }

         //Create ROIs for overlays
         boolean doOverlay = options.getBoolean(Option.OVERLAY);
         if (doOverlay) {
            ImagePlus imp = metadata.getImagePlus();
            boolean isHyperstack = imp.isHyperStack();
            boolean isVisible = imp.isVisible();
            //Compute the time units
            Duration longest = d[nChannels - 1][nSlices - 1][nFrames - 1];
            Time longestUnit = Time.getDimension(longest);
            Time shortestUnit = unit.getNextSmaller();
            shortestUnit = shortestUnit == null ? unit : shortestUnit;
            int digits = String.valueOf((int) unit.toFloat(longest)).length();

            //Overlay fonts (size about 4% of the image, not smaller than 8 pt, monospaced)
            TextRoi.setGlobalJustification(TextRoi.LEFT);
            int h = imp.getHeight()/25;
            h = Math.max(h, 8);
            int offset = h/2;
            Font font = new Font(Font.MONOSPACED, Font.PLAIN, h);

            //Overlay colors (yellow for dark images, dark grey for brightfield)
            Color defaultColor = TextRoi.getColor(); //store it to reset it once done
            Color[] colors = new Color[nChannels];
            int currentC = imp.getC(); //iteration changes the channel!
            if (isVisible) {
               imp.hide();
            }
            for (int i = 0; i < nChannels; i++) {
               imp.setC(i + 1);
               ImageStatistics stats = imp.getStatistics(Measurements.MEDIAN);
               double brightness = (stats.median - stats.min)/(stats.max - stats.min);
               if ((stats.max == 0 && Double.isNaN(brightness) || brightness < 0.5f)) {
                  //CZI reference channels may have a uniform intensity of zero, so no brightness can be computed
                  colors[i] = Color.YELLOW;
               } else {
                  colors[i] = Color.BLUE;
               }
            }
            imp.setC(currentC);

            //Get the image overlay and remove previous timestamps
            Overlay overlay = null;
            if (doOverlay) {
               overlay = imp.getOverlay();
               if (overlay == null) {
                  overlay = new Overlay();
               } else {
                  for (int i = overlay.size() - 1; i >= 0; i--) {
                     Roi roi = overlay.get(i);
                     String value = roi.getProperty(PROPERTY_KEY);
                     if (value != null && value.equals(PROPERTY_VALUE)) {
                        overlay.remove(i);
                     }
                  }
               }
            }

            //Create the ROIs
            switch (report) {
               case FRAME:
                  for (int i = 0; i < nFrames; i++) {
                     String text = Time.format(d[useChannel][useSlice][i], digits, ' ', longestUnit,
                           shortestUnit);
                     addRoi(offset, text, longestUnit, font, colors[0], report, useChannel+1,
                           useSlice+1, i+1, isHyperstack, overlay);
                  }
                  break;
               case CHANNEL:
                  for (int i = 0; i < nChannels; i++) {
                     for (int j = 0; j < nFrames; j++) {
                        String text = Time.format(d[i][useSlice][j], digits,
                              ' ', longestUnit, shortestUnit);
                        addRoi(offset, text, longestUnit, font, colors[i], report,
                              i+1, useSlice+1,
                              j+1, isHyperstack, overlay);
                     }
                  }
                  break;
               case SLICE:
                  for (int i = 0; i < nSlices; i++) {
                     for (int j = 0; j < nFrames; j++) {
                        String text = Time.format(d[useChannel][i][j], digits, ' ', longestUnit,
                              shortestUnit);
                        addRoi(offset, text, longestUnit, font, colors[0],
                              report, useChannel+1, i+1, j+1, isHyperstack, overlay);
                     }
                  }
                  break;
               case ALL:
                  for (int i = 0; i < nChannels; i++) {
                     for (int j = 0; j < nSlices; j++) {
                        for (int k = 0; k < nFrames; k++) {
                           String text = Time.format(d[i][j][k], digits, ' ', longestUnit,
                                 shortestUnit);
                           addRoi(offset, text, longestUnit, font, colors[i], report, i+1, j+1,
                                 k+1, isHyperstack, overlay);
                        }
                     }
                  }
                  break;
            }
            if (doOverlay) {
               imp.setOverlay(overlay);
            }
            TextRoi.setColor(defaultColor); //reset to the original color
            if (isVisible) {
               imp.show();
            }
         }
      }

      /**Returns the median of a list of durations<br><br>
       * <i>Note: </i>the list must be sorted in ascending order
       * @param d the list
       * @return the median duration*/
      private static Duration getMedian(List<Duration> d) {
         if (d.size() % 2 == 0) {
            return (d.get(d.size()/2).plus(d.get(d.size()/2 - 1))).dividedBy(2);
         }
         else {
            return d.get(d.size()/2);
         }
      }

      /**Creates a text ROI and adds it to an overlay and/or to the ROI manager
       * @param offset x/y position of the ROI
       * @param text the text for the ROI
       * @param unit the time unit (is appended in the case of seconds and milliseconds)
       * @param font the text font
       * @param color the stroke color
       * @param report the report option
       * @param channel the channel (starting at 1)
       * @param slice the z slice (starting at 1)
       * @param frame the time point/frame (starting at 1)
       * @param isHyperstack <code>true</code> if the image is a hyperstack
       * @param overlay the overlay or <code>null</code> if the ROI should not be added*/
      private void addRoi(int offset, String text, Time unit, Font font, Color color, Report report,
            int channel, int slice, int frame, boolean isHyperstack, Overlay overlay) {
         if (unit == Time.SECONDS || unit == Time.MILLISECONDS) {
            text = text+" "+unit.SYMBOL;
         }
         TextRoi roi = new TextRoi(offset, offset, text, font);
         roi.setStrokeColor(color);
         StringBuilder sb = new StringBuilder();
         if (report == Report.CHANNEL || report == Report.ALL) {
            sb.append("C");
            sb.append(channel);
            sb.append("-");
         }
         if (report == Report.SLICE || report == Report.ALL) {
            sb.append("Z");
            sb.append(slice);
            sb.append("-");
         }
         sb.append("T");
         sb.append(frame);
         roi.setName(sb.toString());
         if (isHyperstack) {
            roi.setPosition(channel, slice, frame);
         } else {
            roi.setPosition(frame);
         }
         roi.setProperty(PROPERTY_KEY, PROPERTY_VALUE);
         if (overlay != null) {
            overlay.add(roi);
         }
      }
   }
}