package timestamps.parser;

import java.io.BufferedReader;
import java.io.FileReader;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.time.Duration;
import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import timestamps.Timestamps;
import timestamps.shared.Metadata;

/**Timestamp parser for Deltavision Ultra files (.dv)*/
public class Deltavision implements Parser {

   private final static Pattern p =
         Pattern.compile("Extended header (?:Z(\\d+))? (?:W(\\d+))? T(\\d+):timeStampSeconds");

   @Override
   public void parse(Metadata metadata, Timestamps timestamps) throws ParserException {
      int nChannels = metadata.getImagePlus().getNChannels();
      int nSlices = metadata.getImagePlus().getNSlices();
      int nFrames = metadata.getImagePlus().getNFrames();
      List<String> keys = metadata.getKeysBySuffix("timeStampSeconds");
      int total = nChannels * nSlices * nFrames;
      if (total != keys.size()) { //sanity check
         throw new ParserException(getClass(), "Dimension mismatch. Expected "+total+", " +
               "found "+keys.size());
      }
      for (String key : keys) {
         Matcher m = p.matcher(key);
         if (m.find()) {
            String z =  m.group(1);
            String c =  m.group(2);
            String t =  m.group(3);
            if (z == null || c == null || t == null) {
               throw new ParserException(getClass(), "Cannot parse metadata entry "+key);
            }
            float seconds = Float.parseFloat(metadata.get(key).toString());
            long milliseconds = Math.round(seconds * 1000f);
            timestamps.setDuration(Integer.parseInt(c), Integer.parseInt(z),
                  Integer.parseInt(t), Duration.ofMillis(milliseconds));
         } else {
            throw new ParserException(getClass(), "Cannot parse metadata entry "+key);
         }
      }
      //Use the file name as series name
      String fileName = metadata.getOriginalFileInfo().fileName;
      timestamps.setSeriesName(fileName.substring(0, fileName.lastIndexOf(".dv")));
      //Try to get the channel names from the log file
      String log = fileName.replace(".dv", ".dv_log.txt");
      Path path = Paths.get(metadata.getOriginalFileInfo().directory, log);
      if (Files.exists(path) && Files.isReadable(path)) {
         try (BufferedReader br = new BufferedReader(new FileReader(path.toFile()))) {
            String line;
            boolean isRefChannel = fileName.contains("_REF.dv"); //Reference channel
            boolean channelSection = false;
            List<String> channels = new ArrayList<>(nChannels);
            while ((line = br.readLine()) != null) {
               if (line.equals("Channels:")) {
                  channelSection = true;
               } else if (channelSection) {
                  if (line.isEmpty()) {
                     break;
                  } else {
                     String channel = line.substring(0, line.indexOf("EX:") - 1).trim();
                     if (line.contains("Reference Channel")) {
                        if (isRefChannel) {
                           timestamps.setChannelName(0, channel);
                        }
                     } else {
                        channels.add(channel);
                     }
                  }
               }
            }
            if (!isRefChannel && !channels.isEmpty()) {
               for (int i = 0; i < channels.size(); i++) {
                  timestamps.setChannelName(i, channels.get(i));
               }
            }
         } catch (Exception e) {
            //ignore (default channel names can be used instead)
         }
      }
   }
}
