package timestamps.parser;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.time.Duration;
import loci.common.services.DependencyException;
import loci.common.services.ServiceException;
import loci.common.services.ServiceFactory;
import loci.formats.FormatException;
import loci.formats.ImageReader;
import loci.formats.meta.IMetadata;
import loci.formats.services.OMEXMLService;
import ome.units.UNITS;
import ome.units.quantity.Time;
import timestamps.Timestamps;
import timestamps.shared.Metadata;

/**Format-independent reader utilizing the Bio-Formats OME-XML metadata model. Based on
 * https://github.com/ome/bioformats/blob/v6.5.0/components/formats-gpl/utils/PrintTimestamps.java*/
public class OMEXML implements Parser {

   @Override
   public void parse(Metadata metadata, Timestamps timestamps) throws ParserException {
      String dir = metadata.getOriginalFileInfo().directory;
      String fileName = metadata.getOriginalFileInfo().fileName;
      String shortName = fileName.substring(0, fileName.lastIndexOf("."));
      String extension = fileName.substring(fileName.lastIndexOf("."));
      if (fileName.endsWith(".ome.tif")) {
         shortName = shortName.substring(0, shortName.lastIndexOf("."));
         extension = ".ome.tif";
      }
      Path path = Paths.get(dir, fileName);
      if (!Files.exists(path) || !Files.isReadable(path)) {
         throw new ParserException(getClass(),"Cannot find or read "+path);
      }
      ImageReader reader = new ImageReader();
      IMetadata bfMeta;
      //noinspection TryFinallyCanBeTryWithResources
      try {
         ServiceFactory factory = new ServiceFactory();
         OMEXMLService service = factory.getInstance(OMEXMLService.class);
         bfMeta = service.createOMEXMLMetadata();
         reader.setMetadataStore(bfMeta);
         reader.setId(path.toString());
         //Get the current series and its name
         int series = 0;
         int nSeries = reader.getSeriesCount();
         if (nSeries == 1) {
            timestamps.setSeriesName(shortName);
         } else {
            //Figure out the current series
            for (int i = 0; i < nSeries; i++) {
               String name = bfMeta.getImageName(i);
               if (metadata.getTitle().contains(name)) { //series names are appended to the title
                  series = i;
                  String seriesName = name.replace(extension, "");
                  seriesName = seriesName.replaceAll("#", "");
                  seriesName = seriesName.replaceAll(" ", "_");
                  timestamps.setSeriesName(seriesName);
               }
            }
         }
         int planeCount = bfMeta.getPlaneCount(series);
         for (int i = 0; i < planeCount; i++) {
            Time deltaT = bfMeta.getPlaneDeltaT(series, i);
            if (deltaT == null) {
               continue;
            }
            int c = bfMeta.getPlaneTheC(series, i).getValue();
            int z = bfMeta.getPlaneTheZ(series, i).getValue();
            int t = bfMeta.getPlaneTheT(series, i).getValue();
            Duration d = Duration.ofNanos(deltaT.value(UNITS.NANOSECOND).longValue());
            timestamps.setDuration(c, z, t, d);
         }
      } catch (IOException ioe) {
         throw new ParserException(getClass(), "IO error in Bio-Formats reader", ioe);
      } catch (DependencyException de) {
         throw new ParserException(getClass(), "Dependency error in Bio-Formats reader", de);
      } catch (FormatException fe) {
         throw new ParserException(getClass(), "Format error in Bio-Formats reader", fe);
      } catch (ServiceException se) {
         throw new ParserException(getClass(), "Metadata service error in Bio-Formats reader", se);
      } finally {
         try {
            reader.close();
         } catch (IOException e) {
            //ignore;
         }
      }
   }
}