package timestamps.parser;

import java.io.File;
import java.time.Duration;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.Hashtable;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import loci.formats.in.MetamorphReader;
import timestamps.Timestamps;
import timestamps.shared.Metadata;
import timestamps.shared.Strings;

/**Timestamp parser for Metamorph/VisiView files (*.nd/stk/tif)*/
public class VisiView implements Parser {

   private static final String DEFAULT_CHANNEL_NAME = "CH1";
   private static final Comparator<String> STRING_COMPARATOR = Strings.getNaturalComparator();
   private static final DateTimeFormatter DATE_TIME_FORMATTER =
         DateTimeFormatter.ofPattern("dd/MM/yyyy HH:mm:ss:SSS");

   private final MetamorphReader reader;
   private Metadata metaData;
   private Timestamps timestamps;
   private String baseName;
   private int nChannels;
   private int nSlices;
   private int nFrames;
   private LocalDateTime[][][] dateTimes;

   //Fields populated from an .nd file. If none is present, they will always be false
   private boolean doWave;
   private boolean doTimelapse;

   /**Constructor*/
   public VisiView() {
      this.reader = new MetamorphReader();
      reader.setOriginalMetadataPopulated(true);
      reader.setGroupFiles(false);
   }

   /**{@inheritDoc}*/
   @Override
   public void parse(Metadata metadata, Timestamps timestamps) throws ParserException {
      this.metaData = metadata;
      this.timestamps = timestamps;
      this.doWave = metadata.getBoolean("DoWave");
      boolean doZSeries = metadata.getBoolean("DoZSeries");
      this.doTimelapse = metadata.getBoolean("DoTimelapse");
      boolean doStage = metadata.getBoolean("DoStage");
      nChannels = metadata.getImagePlus().getNChannels();
      nSlices = metadata.getImagePlus().getNSlices();
      nFrames = metadata.getImagePlus().getNFrames();
      dateTimes = new LocalDateTime[nChannels][nSlices][nFrames];

      //Simple case: multi-frame, single channel, single position STK. We can trust the metadata.
      if (nFrames > 1 && !doWave && !doZSeries && !doStage) {
         String illuminationSetting = metaData.getString("Channel #0 _IllumSetting_");
         if (illuminationSetting != null && !illuminationSetting.isEmpty()) {
            timestamps.setChannelName(0, illuminationSetting);
         } else {
            timestamps.setChannelName(0, DEFAULT_CHANNEL_NAME);
         }
         List<String> dates = metadata.getValuesForKeys("creationDate");
         List<String> times = metadata.getValuesForKeys("creationTime");
         if (dates.size() != nFrames || times.size() != nFrames) {
            throw new ParserException(getClass(), "Frame number does not match metadata");
         }
         StringBuilder sb;
         for (int i = 0; i < nFrames; i++) {
            sb = new StringBuilder(dates.get(i));
            sb.append(" ");
            sb.append(times.get(i));
            LocalDateTime dateTime = LocalDateTime.parse(sb.toString(), DATE_TIME_FORMATTER);
            dateTimes[0][0][i] = dateTime;
         }
         reportDurations();
      }

      /*All other images may have wrong InfoProperties and FileInfos, and titles originating from the first
      * series image or the one dragged onto ImageJ (or the Bio-Formats shortcut window), not necessarily
      * from the one finally selected in Bio-Formats series selection dialog!*/

      //Get the base name from the FileInfo
      String originalFileName = metadata.getOriginalFileInfo().fileName;
      if (originalFileName.endsWith(".nd")) {
         baseName = originalFileName.substring(0, originalFileName.length() - 3);
      } else {
         StringBuilder sb = new StringBuilder("^(?:(.+\\d+)_)");
         if (doWave) {
            sb.append("w\\d.+");
         } else {
            sb.append(".+");
         }
         if (doStage) {
            sb.append("_s\\d+");
         }
         if (metadata.getImagePlus().getNFrames() > 1) { //using DoTimelapse is always false without .nd file
            sb.append("_t\\d+");
         }
         sb.append("(?:\\.tif|\\.stk)$");
         Matcher matcher = Pattern.compile(sb.toString()).matcher(originalFileName);
         if (matcher.find()) {
            baseName = matcher.group(1);
         }
         if (baseName == null) {
            throw new ParserException(getClass(), "Cannot retrieve the base name from "
                  + originalFileName);
         }
      }

      //Get channel (wavelength) names from the info property
      String[] allChannels = null;
      if (doWave) {
         List<String> list = metadata.getValuesForKeys("WaveName");
         //noinspection ToArrayCallWithZeroLengthArrayArgument
         allChannels = list.toArray(new String[list.size()]);
      }

      //Check the info property if the image is a series of a multiposition experiment
      if (doStage) {
         // Since it's from a series, we cannot trust the FileInfo or InfoProperty. Inspect the title.
         Pattern stageSeriesPattern =
               Pattern.compile(".+\\.(?:nd|stk|tif)\\s-\\sStage(\\d)\\s\"(.+)\"(?:;\\s(.+))?");
         Matcher matcher = stageSeriesPattern.matcher(metadata.getTitle());
         if (matcher.matches()) {
            String positionString = matcher.group(1);
            if (positionString == null) {
               throw new ParserException(getClass(), "Cannot parse the image title of"
                     + metadata.getTitle() + " (stage position not found)");
            }
            int positionIndex = Integer.parseInt(positionString);
            String positionLabel = matcher.group(2);
            String channel = matcher.group(3);
            if (channel == null) {
               //All channels contained in image, no separate series to be sorted out
               parsePosition(positionIndex, positionLabel, true, allChannels);
            } else {
               if (channel.contains("/")) { //channel separator in the title
                  //Two or more channels in this series, additional channel(s) in another
                  parsePosition(positionIndex, positionLabel, false, channel.split("/"));
               } else {
                  //One channel in this series, additional channel(s) in another
                  parsePosition(positionIndex, positionLabel, false, channel);
               }
            }
         } else {
            //Something is wrong with the title
            throw new ParserException(getClass(), "Cannot retrieve stage positions from "
                  + metadata.getTitle());
         }
      } else {
         /* Image not from a multiposition series. Check if all channels have the same z-dimension.
          * If not, channels with z-stacks end up in one image, single-sliced reference images in onother */
         if (doZSeries && doWave) {
            List<String> waveDoZList = metadata.getValuesForKeys("WaveDoZ");
            if (waveDoZList.stream().allMatch(waveDoZList.get(0)::equals)) {
               //All channels have the same z-dimension and should, hence, be present in the ImagePlus
               parsePosition(null, null, true, allChannels);
            } else {
               //Inspect the title for channels
               Pattern channelPattern =
                     Pattern.compile(".+\\.(?:nd|stk|tif)\\s-\\s(.+)");
               Matcher matcher = channelPattern.matcher(metadata.getTitle());
               if (matcher.matches()) {
                  String channel = matcher.group(1);
                  if (channel.contains("/")) {
                     //Two or more channels in this series, additional channel(s) in another one
                     parsePosition(null, null, false,
                           channel.split("/"));
                  } else {
                     //One channel in this series, additional channel(s) in another one
                     parsePosition(null, null, false, channel);
                  }
               } else {
                  throw new ParserException(getClass(), "Cannot retrieve channels from title"
                              + metadata.getTitle());
               }
            }
         } else {
            //All channels have the same z-dimension and should be present, or there is only one channel
            parsePosition(null, null, true, allChannels);
         }
      }
      reportDurations();
   }

   /**Stores channel names in the metadata and retrieves image files comprising the stage
    * position and channel(s) provided
    * @param positionIndex the index of the stage position or <code>null</code> for single-position images
    * @param positionLabel the stage position label or <code>null</code> for single-position images
    * @param allChannelsInImage <code>true</code> if there is no separate series with another channel
    *     (e.g. a single DIC reference slice accompanying a fluorescence stack)
    * @param channels the channel names (in the order as acquired)*/
   private void parsePosition(Integer positionIndex, String positionLabel,
         boolean allChannelsInImage, String... channels) throws ParserException {
      StringBuilder seriesNameBuilder = new StringBuilder(baseName);
      if (channels == null) {
         //Single-channel image
         String channelName = metaData.getString("ImageName");
         if (channelName == null) {
            channelName = DEFAULT_CHANNEL_NAME;
         }
         channels = new String[]{channelName};
      }
      File dir = new File(metaData.getOriginalFileInfo().directory);
      String[] filesInDirectory = dir.list();
      if (filesInDirectory == null || filesInDirectory.length == 0) {
         throw new ParserException(getClass(), "Cannot read "+dir);
      }
      for (int i = 0; i < channels.length; i++) {
         timestamps.setChannelName(i, channels[i]);
         List<String> usedFiles = new ArrayList<>();
         StringBuilder fileNameBuilder = new StringBuilder("^");
         fileNameBuilder.append(baseName);
         fileNameBuilder.append("_");
         if (doWave) {
            fileNameBuilder.append("w\\d");
         }
         fileNameBuilder.append(channels[i]);
         if (positionIndex != null) {
            fileNameBuilder.append("_s");
            fileNameBuilder.append(positionIndex);
         }
         if (doTimelapse) {
            fileNameBuilder.append("_t\\d+");
         }
         fileNameBuilder.append("\\.(?:tif|stk)$");
         Pattern p = Pattern.compile(fileNameBuilder.toString());
         Matcher m;
         for (String fileName : filesInDirectory) {
            m = p.matcher(fileName);
            if (m.matches()) {
               usedFiles.add(fileName);
            }
         }
         usedFiles.sort(STRING_COMPARATOR);
         /* Sometimes there is one more single TIFF, which is not included in the ND file, hence not
          * read by Bio-Formats, and thus the list of files may be larger than the frame number, resulting in
          * downstream array-out-of-bounds exception. For safety, use the smaller of the two values.*/
         for (int j = 0; j < Math.min(usedFiles.size(), nFrames); j++) {
            readSingleFile(j, i, usedFiles.get(j));
         }
      }
      //Complete the series name
      if (!allChannelsInImage && doWave) {
         for (int i = 0; i < channels.length; i++) {
            seriesNameBuilder.append("_w");
            int channelNumber = i + 1;
            seriesNameBuilder.append(channelNumber);
            seriesNameBuilder.append(channels[i]);
         }
      } else if (!doWave) {
         seriesNameBuilder.append("_");
         seriesNameBuilder.append(channels[0]);
      }
      if (positionIndex != null) {
         seriesNameBuilder.append("_s");
         seriesNameBuilder.append(positionIndex);
         seriesNameBuilder.append("_");
         seriesNameBuilder.append(positionLabel);
      }
      timestamps.setSeriesName(seriesNameBuilder.toString());
   }

   /**Reads a single TIFF or STK file<p>
    * <strong>Note:</strong> multi-frame, single-slice STKs are handled above, not within this method
    * @param frame time point/frame index
    * @param channel the channel
    * @param fileName the name of the file*/
   private void readSingleFile(int frame, int channel, String fileName) throws ParserException {
      try {
         reader.setId(metaData.getOriginalFileInfo().directory+fileName);
      } catch (Exception e) {
         throw new ParserException(getClass(), "Error while reading "+fileName);
      }
      /* ImageReader#getSeriesMetadataValue(key) throws a null pointer exception. Fetch the hashtable
       * and query this one rather than by the Bio-Formats API*/
      Hashtable<String, Object> seriesMetadata = reader.getSeriesMetadata();
      int slices = reader.getSizeZ();
      if (slices == 1) {
         //Single-plane TIFF. Stores only one entry 'DateTime'
         String creationTime = seriesMetadata.get("DateTime").toString();
         LocalDateTime dateTime = LocalDateTime.parse(creationTime, DATE_TIME_FORMATTER);
         dateTimes[channel][0][frame] = dateTime;
      } else {
         /* STKs store separate fields for creation date and time. Indices in the keys are padded
          * with zeros up to the digits of the largest index. Avoid having to code this by querying
          * the hashtable for the common prefix and sort the values alphanumerically afterwards*/
         List<String> dates = new ArrayList<>(slices);
         List<String> times = new ArrayList<>(slices);
         for (String key : seriesMetadata.keySet()) {
            if (key.startsWith("creationDate")) {
               dates.add(seriesMetadata.get(key).toString());
            } else if (key.startsWith("creationTime")) {
               times.add(seriesMetadata.get(key).toString());
            }
         }
         dates.sort(STRING_COMPARATOR);
         times.sort(STRING_COMPARATOR);
         //Just one check that the numbers match
         if (dates.size() != slices || times.size() != slices) {
            throw new ParserException(getClass(), "Slice number does not match metadata");
         }
         StringBuilder sb;
         for (int i = 0; i < slices; i++) {
            sb = new StringBuilder(dates.get(i));
            sb.append(" ");
            sb.append(times.get(i));
            LocalDateTime dateTime = LocalDateTime.parse(sb.toString(), DATE_TIME_FORMATTER);
            dateTimes[channel][i][frame] = dateTime;
         }
      }
      try {
         reader.close();
      } catch (Exception e) {
         throw new ParserException(getClass(), e.getMessage(), e);
      }

   }

   /**Computes durations and reports them to the {@link Timestamps} class*/
   private void reportDurations() {
      LocalDateTime first = dateTimes[0][0][0];
      for (int i = 0; i < nChannels; i++) {
         for (int j = 0; j < nSlices; j++) {
            for (int k = 0; k < nFrames; k++) {
               timestamps.setDuration(i, j, k, Duration.between(first, dateTimes[i][j][k]));
            }
         }
      }
   }
}
