package timestamps.parser;

import java.io.IOException;
import java.time.Duration;
import java.util.ArrayList;
import java.util.List;
import loci.formats.FormatException;
import loci.formats.ImageReader;
import timestamps.Timestamps;
import timestamps.shared.Metadata;

/**Timestamp parser for Olympus cellSens Dimension files (.vsi)*/
public class cellSens implements Parser {

   /**{@inheritDoc}*/
   @Override
   public void parse(Metadata metadata, Timestamps timestamps) throws ParserException {
      /* We need to know if the image is from a series and, if so, get the series count and current name.
       * Note that keys of the info property depend on the way series are opened: if a user selects more
       * than series, the series name is prepended to all keys. It's hard to figure that out. As a
       * workaround, we try to get this info by querying for a key which exists for sure (DimensionOrder)*/
      String title = metadata.getTitle();
      int seriesCount;
      String keyPrefix = "";
      if (!metadata.containsKey("DimensionOrder")) {
         keyPrefix = title.substring(title.indexOf(".vsi -") + 7)+" ";
         System.out.println("KEY PREFIX: "+keyPrefix);
      }
      String directory = metadata.getOriginalFileInfo().directory;
      String fileName = metadata.getOriginalFileInfo().fileName;
      try (ImageReader reader = new ImageReader()) {
         reader.setId(directory+fileName);
         seriesCount = reader.getSeriesCount();
      } catch (IOException | FormatException e) {
         throw new ParserException(getClass(), "Error while reading "+fileName+" ("
               +e.getMessage()+")");
      }
      if (seriesCount > 2 && title.contains(".vsi - ")) {
         //A series (the macro image also counts as series)
         String seriesName = title.replace(".vsi - ", "_");
         timestamps.setSeriesName(seriesName.replace(", ", "_"));
      } else if (title.endsWith(".vsi")) {
         timestamps.setSeriesName(title.substring(0, title.lastIndexOf(".vsi")));
      } else {
         timestamps.setSeriesName(fileName.substring(0, fileName.lastIndexOf(".vsi")));
      }

      //Channel names
      List<String> channelNames = metadata.getValuesForKeys(keyPrefix+"Channel name #");
      for (int i = 0; i < channelNames.size(); i++) {
         timestamps.setChannelName(i, channelNames.get(i));
      }
      int nChannels = metadata.getImagePlus().getNChannels();
      int nSlices = metadata.getImagePlus().getNSlices();
      int nFrames = metadata.getImagePlus().getNFrames();

      //Get a list of time values
      List<String> units = metadata.getValuesForKeys(keyPrefix+"Units #");
      List<String> values = metadata.getKeys(keyPrefix+"Value #");
      if (units.size() == values.size() + 1) {
         /*Sometimes, there is a strange extra entry "Units #1 = 10^-3s^1", without a value.
          *The true entry for time point #1 is then actually "Units #001".*/
         if (metadata.containsKey("Units #1") && metadata.containsKey("Units #001")) {
            metadata.remove("Units #1");
            //We have to read the data again
            units = metadata.getValuesForKeys(keyPrefix+"Units #");
         }
      }
      List<Duration> times = new ArrayList<>();
      for (int i = 0; i < units.size(); i++) {
         if (units.get(i).equals("10^-3s^1")) { //milliseconds
            int millisRounded = Math.round(Float.parseFloat(metadata.get(values.get(i)).toString()));
            times.add(Duration.ofMillis(millisRounded));
         }
      }
      int total = nChannels * nSlices * nFrames;
      if (total > times.size()) { //sanity check
         throw new ParserException(getClass(), "Dimension mismatch. Expected "+total+", " +
               "found "+times.size());
      }

      //Report the timestamps
      for (int i = 0; i < nChannels; i++) {
         for (int j = 0; j < nSlices; j++) {
            for (int k = 0; k < nFrames; k++) {
               int n = i * nSlices * nFrames + j * nSlices + k;
               timestamps.setDuration(i, j, k, times.get(n));
            }
         }
      }
   }
}
