package timestamps.shared;

import ij.IJ;
import java.awt.Component;
import java.awt.HeadlessException;
import java.awt.Image;
import java.io.File;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.filechooser.FileNameExtensionFilter;

/**Extension of {@link JFileChooser} dialogs displaying the ImageJ icon. With static convenience methods
 * to generate directory choosers and file save dialogs.*/
@SuppressWarnings("serial")
public class IJFileChooser extends JFileChooser {

   /**Helper method to set the ImageJ icon.
    * @param parent the parent component
    * @return the dialog*/
   @Override
   protected JDialog createDialog(Component parent) throws HeadlessException {
      JDialog dialog = super.createDialog(parent);
      Image icon = IJ.getInstance().getIconImage();
      if (icon != null) {
         dialog.setIconImage(IJ.getInstance().getIconImage());
      }
      return dialog;
   }

   /**Displays a directory chooser.
    * @return the directory chosen
    * @see IJFileChooser#getDirectory(File)
    * @see IJFileChooser#getDirectory(String)
    * @see IJFileChooser#getDirectory(String, File)*/
   @SuppressWarnings("unused")
   public static File getDirectory() {
      return getDirectory("Select a directory...", new File(System.getProperty("user.home")));
   }

   /**Displays a directory chooser with the titel provided.
    * @param title the dialog title
    * @return the directory chosen
    * @see IJFileChooser#getDirectory()
    * @see IJFileChooser#getDirectory(File)
    * @see IJFileChooser#getDirectory(String, File)*/
   @SuppressWarnings("unused")
   public static File getDirectory(String title) {
      return getDirectory(title, new File(System.getProperty("user.home")));
   }

   /**Displays a directory chooser with the initial directory provided.
    * @param initial the initial directory
    * @return the directory chosen
    * @see IJFileChooser#getDirectory()
    * @see IJFileChooser#getDirectory(String)
    * @see IJFileChooser#getDirectory(String, File)*/
   @SuppressWarnings("unused")
   public static File getDirectory(File initial) {
      return getDirectory("Select a directory...", initial);
   }

   /**Displays a directory chooser with the initial directory and title provided.
    * @param title the dialog title
    * @param initial the initial directory
    * @return the directory chosen
    * @see IJFileChooser#getDirectory()
    * @see IJFileChooser#getDirectory(String)
    * @see IJFileChooser#getDirectory(File)*/
   @SuppressWarnings("unused")
   public static File getDirectory(String title, File initial) {
      IJFileChooser fc = new IJFileChooser();
      fc.setDialogTitle(title);
      fc.setCurrentDirectory(initial);
      fc.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
      fc.setMultiSelectionEnabled(false);
      int value = fc.showOpenDialog(null);
      if (value == JFileChooser.APPROVE_OPTION) {
         return fc.getSelectedFile();
      }
      return null;
   }

   /**Displays a file save dialog with initial directory or file.
    * @param initial the suggested file or an initial directory
    * @return the file chosen or <code>null</code> if the user cancelled
    * @see IJFileChooser#saveFile(File, String, String[])
    * @see IJFileChooser#saveFile(String, File, String, String[])*/
   @SuppressWarnings("unused")
   public static File saveFile(File initial) {
      return saveFile("Select a file name", initial, null, null);
   }

   /**Displays a file save dialog with initial directory or file and file name filter.
    * @param initial the suggested file or an initial directory
    * @param description file description (e.g., "Text files", "XML files",...)
    * @param extensions suggested file extensions (e.g., "txt", "log"...)
    * @return the file chosen or <code>null</code> if the user cancelled
    * @see IJFileChooser#saveFile(File)
    * @see IJFileChooser#saveFile(String, File, String, String[])*/
   @SuppressWarnings("unused")
   public static File saveFile(File initial, String description, String[] extensions) {
      return saveFile("Select a file name", initial, description, extensions);
   }

   /**Displays a file save dialog with initial directory or file, the dialog title provided and file name
    * filter.
    * @param title the title of the dialog
    * @param initial the suggested file or an initial directory
    * @param description file description (e.g., "Text files", "XML files",...)
    * @param extensions suggested file extensions (e.g., "txt", "log"...)
    * @return the file chosen or <code>null</code> if the user cancelled
    * @see IJFileChooser#saveFile(File)
    * @see IJFileChooser#saveFile(File, String, String[])*/
   @SuppressWarnings("unused")
   public static File saveFile(String title, File initial, String description,
         String[] extensions) {
      IJFileChooser fc = new IJFileChooser();
      fc.setDialogTitle(title);
      fc.setFileSelectionMode(JFileChooser.FILES_ONLY);
      fc.setMultiSelectionEnabled(false);
      if (initial.isDirectory()) {
         fc.setCurrentDirectory(initial);
      } else {
         fc.setSelectedFile(initial);
      }
      if (description != null && extensions != null && extensions.length > 0) {
         fc.setFileFilter(new FileNameExtensionFilter(description, extensions));
      }
      int value = fc.showSaveDialog(null);
      if (value == JFileChooser.APPROVE_OPTION) {
         File f = fc.getSelectedFile();
         if (f.getParentFile().canWrite()) {
            return f;
         }
      }
      return null;
   }
}
