package timestamps.shared;

import ij.IJ;
import ij.ImageJ;
import ij.Menus;
import ij.WindowManager;
import ij.plugin.PlugIn;
import ij.plugin.frame.PlugInFrame;
import ij.plugin.frame.Recorder;
import java.awt.Frame;
import java.awt.Image;
import java.awt.event.ActionEvent;
import java.awt.event.FocusAdapter;
import java.awt.event.FocusEvent;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import javax.swing.AbstractAction;
import javax.swing.Box;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.KeyStroke;

/**Modification of the original {@link PlugInFrame} from ImageJ implementing a Swing {@link JFrame}
 * rather then AWT {@link Frame}, adapters rather than listeners, and a menu bar.<br><br>
 * <i>Note:</i> subclasses must implement {@link PlugIn#run(String).}*/
@SuppressWarnings("serial")
public abstract class IJFrame extends JFrame {

   protected ImageJ ij;
   protected JMenuBar menuBar;

   /**Constructor.
    * @param title the title to be displayed*/
   @SuppressWarnings("unused")
   public IJFrame(String title) {
      ij = IJ.getInstance();
      Image icon = ij.getIconImage();
      if (icon != null) {
         setIconImage(icon);
      }
      setTitle(title);
      setLocationRelativeTo(null);
      setDefaultCloseOperation(DO_NOTHING_ON_CLOSE);
      WindowManager.addWindow(this);
      addWindowListener(new WindowAdapter() {

         public void windowClosing(WindowEvent we) {
            if (we.getSource() == IJFrame.this) {
               quit();
               if (ij != null && Recorder.record) {
                  Recorder.record("run", "Close");
               }
            }
         }

         public void windowActivated(WindowEvent we) {
            if (ij != null && IJ.isMacintosh()) {
               IJ.wait(10); //May be needed for Java 1.4 on OS X
               setMenuBar(Menus.getMenuBar());
            }
            WindowManager.setWindow(IJFrame.this);
         }
      });
      addFocusListener(new FocusAdapter() {

         public void focusGained(FocusEvent e) {
            WindowManager.setWindow(IJFrame.this);
         }
      });
      menuBar = new JMenuBar();
      setJMenuBar(menuBar);
      //File menu
      JMenu fileMenu = new JMenu("File");
      fileMenu.setMnemonic(KeyEvent.VK_F);
      menuBar.add(fileMenu);
      JMenuItem exitItem = new JMenuItem();
      exitItem.setAction(new ActionExit());
      exitItem.setToolTipText("Quits the plugin");
      fileMenu.add(exitItem);
      //Options
      JMenu optionsMenu = new JMenu("Options");
      optionsMenu.setMnemonic(KeyEvent.VK_O);
      menuBar.add(optionsMenu);
      JCheckBoxMenuItem onTopItem = new JCheckBoxMenuItem();
      onTopItem.setAction(new ActionOnTop());
      onTopItem.setToolTipText("Keeps the plugin in the desktop foreground");
      optionsMenu.add(onTopItem);
      menuBar.add(Box.createHorizontalGlue());
      //Help
      JMenu helpMenu = new JMenu("?");
      helpMenu.setMnemonic(KeyEvent.VK_H);
      menuBar.add(helpMenu);
      JMenuItem aboutItem = new JMenuItem();
      aboutItem.setAction(new ActionAbout());
      helpMenu.add(aboutItem);
   }

   /**Subclasses should display an 'About' dialog when called.*/
   protected abstract void displayAbout();

   /**Convenience method to get a menubar item.
    * @param menu the name of the menu
    * @return the menu or <code>null</code> if none with the provided name is found*/
   @SuppressWarnings("unused")
   protected JMenu getMenu(@SuppressWarnings("SameParameterValue") String menu) {
      for (int i = 0; i < menuBar.getMenuCount(); i++) {
         JMenu item = menuBar.getMenu(i);
         if (item.getText().equals(menu)) {
            return item;
         }
      }
      return null;
   }

   /**Closes this window. Subclasses can override this method to execute other
    * operations before exiting.*/
   protected void quit() {
      WindowManager.removeWindow(this);
      dispose();
   }

   //Menu actions

   protected class ActionAbout extends AbstractAction {

      ActionAbout() {
         super("About...");
         putValue(ACTION_COMMAND_KEY, "ABOUT_CMD");
         putValue(ACCELERATOR_KEY, KeyStroke.getKeyStroke(KeyEvent.VK_A, InputEvent.ALT_DOWN_MASK));
      }

      @Override
      public void actionPerformed(ActionEvent evt) {
         IJFrame.this.displayAbout();
      }
   }

   protected class ActionExit extends AbstractAction {

      ActionExit() {
         super("Exit");
         putValue(ACTION_COMMAND_KEY, "EXIT_CMD");
         putValue(ACCELERATOR_KEY, KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, InputEvent.ALT_DOWN_MASK));
      }

      @Override
      public void actionPerformed(ActionEvent evt) {
         IJFrame.this.quit();
      }
   }

   protected class ActionOnTop extends AbstractAction {

      ActionOnTop() {
         super("Always on top");
         putValue(ACTION_COMMAND_KEY, "ON_TOP_CMD");
         putValue(ACCELERATOR_KEY, KeyStroke.getKeyStroke(KeyEvent.VK_T, InputEvent.ALT_MASK));
         putValue(SELECTED_KEY, false);
      }

      @Override
      public void actionPerformed(ActionEvent evt) {
         IJFrame.this.setAlwaysOnTop((boolean) this.getValue(SELECTED_KEY));
      }
   }
}
