package timestamps.shared;

import ij.ImagePlus;
import ij.io.FileInfo;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Set;

/**Helper class parsing an {@link ImagePlus} info property into key/value pairs. Lines containing an equals
 *  sign are considered as a pair. Also provides access to the image and its {@link FileInfo} metadata.<br>
 * <br><i>Note:</i> in many instances, one does not need this class, as Bio-Formats metadata are
 * accessible by {@link ImagePlus#getNumericProperty(String)} and
 * {@link ImagePlus#getStringProperty(String)}. However, parsing into a {@link LinkedHashMap} allows to
 * search in a strict order, making it much more convenient to handle indexed keys or values (e.g.,
 * Value[0], Value[1], Unit[00], Unit[01] in Olympus files)*/
public class Metadata {

   private final ImagePlus imp;
   private final LinkedHashMap<String, String> map;

   /**Constructor.
    * @param imp the image*/
   @SuppressWarnings("unused")
   public Metadata(ImagePlus imp) {
      this.imp = imp;
      map = new LinkedHashMap<>();
      String[] info = imp.getInfoProperty().split("\r\n|[\r\n]");
      for (String line : info) {
         String[] pair = line.split(" = ");
         if (pair.length == 2) {
            String value = pair[1].trim();
            //Remove leading and trailing quotes
            if (value.startsWith("\"") && value.endsWith("\"")) {
               value = value.substring(1, value.length() - 1);
            }
            map.put(pair[0].trim(), value);
         }
      }
   }

   /**Returns the {@link ImagePlus} underlying this metadata object.
    * @return the image*/
   @SuppressWarnings("unused")
   public ImagePlus getImagePlus() {
      return imp;
   }

   /**Returns the title of the image.
    * @return the image*/
   @SuppressWarnings("unused")
   public String getTitle() {
      return imp.getTitle();
   }

   /**Returns the original {@link FileInfo} associated of the {@link ImagePlus}.
     * @return the {@link FileInfo}*/
   @SuppressWarnings("unused")
   public FileInfo getOriginalFileInfo() {
      return imp.getOriginalFileInfo();
   }

   /**Searches through keys of the image properties (ImageJ menu command 'Image' -> 'Show Info...'), whereby
    * a line is considered as a key/value pair if it contains an equals sign.
    * @param key the key
    * @return <code>true</code> or <code>false</code>*/
   @SuppressWarnings("unused")
   public boolean containsKey(String key) {
      return map.containsKey(key);
   }

   /**Searches through values of the image properties (ImageJ menu command 'Image' -> 'Show Info...'), whereby
    * a line is considered as a key/value pair if it contains an equals sign.
    * @param value the value
    * @return <code>true</code> or <code>false</code>*/
   @SuppressWarnings("unused")
   public boolean containsValue(String value) {
      return map.containsValue(value);
   }

   /**Returns the value associated with the given key.
    * @param key the key
    * @return the value or <code>null</code>*/
   @SuppressWarnings("unused")
   public Object get(String key) {
      return map.get(key);
   }

   /**Tries to parse a boolean for a value in the properties map.
    * @param key the key in the map
    * @return <code>true</code> or <code>false</code>.<br>
    *    <strong>Note:</strong> returns also <code>false</code> if the key is not present*/
   @SuppressWarnings("unused")
   public boolean getBoolean(String key) {
      String s = map.get(key);
      if (s == null || s.isEmpty()) {
         return false;
      }
      return Boolean.parseBoolean(s);
   }

   /**Returns the value associated with the given key as {@link String} representation.
    * @param key the key
    * @return the value or an empty string (but never <code>null</code>)*/
   @SuppressWarnings("unused")
   public String getString(String key) {
      Object value = map.get(key);
      if (value == null) {
         return "";
      }
      return value.toString();
   }

   /**Returns a list of keys sharing a common prefix (e.g., "Stage1", "Stage2"...).
    * The order of items corresponds to that in the original {@link ImagePlus} info property.<br>
    * <i>Note:</i> leading and trailing quotes are removed
    * @param prefix the prefix
    * @return the list (empty if no key with the prefix is present)
    * @see Metadata#getKeysBySuffix(String)*/
   @SuppressWarnings("unused")
   public List<String> getKeys(String prefix) {
      List<String> list = new ArrayList<>();
      for (String key : map.keySet()) {
         if (key.startsWith(prefix)) {
            list.add(key);
         }
      }
      return list;
   }

   /**Returns a list of keys sharing a common suffix (e.g., "XX_Time", "XY_Time"...).
    * The order of items corresponds to that in the original {@link ImagePlus} info property.<br>
    * <i>Note:</i> leading and trailing quotes are removed
    * @param suffix the prefix
    * @return the list (empty if no key with the prefix is present)
    * @see Metadata#getKeys(String)*/
   @SuppressWarnings("unused")
   public List<String> getKeysBySuffix(String suffix) {
      List<String> list = new ArrayList<>();
      for (String key : map.keySet()) {
         if (key.endsWith(suffix)) {
            list.add(key);
         }
      }
      return list;
   }

   /**Returns a list of values of keys sharing a common prefix (e.g., "Stage1", "Stage2"...).
    * The order of items corresponds to that in the original {@link ImagePlus} info property.<br>
    * <i>Note:</i> leading and trailing quotes are removed.
    * @param prefix the prefix
    * @return the list (empty if no key with the prefix is present)
    * @see Metadata#getValuesForKeysBySuffix(String)*/
   @SuppressWarnings("unused")
   public List<String> getValuesForKeys(String prefix) {
      List<String> list = new ArrayList<>();
      for (String key : map.keySet()) {
         if (key.startsWith(prefix)) {
            list.add(map.get(key));
         }
      }
      return list;
   }

   /**Returns a list of values of keys sharing a common suffix (e.g., "XX_Time", "XY_Time"...).
    * The order of items corresponds to that in the original {@link ImagePlus} info property.<br>
    * <i>Note:</i> leading and trailing quotes are removed.
    * @param suffix the suffix
    * @return the list (empty if no key with the suffix is present)
    * @see Metadata#getValuesForKeys(String)*/
   @SuppressWarnings("unused")
   public List<String> getValuesForKeysBySuffix(String suffix) {
      List<String> list = new ArrayList<>();
      for (String key : map.keySet()) {
         if (key.endsWith(suffix)) {
            list.add(map.get(key));
         }
      }
      return list;
   }

   /**Removes a key
    * @param key the key
    * @return the value associated previously with the key (or <code>null</code>)*/
   public String remove(String key) {
      return map.remove(key);
   }

   /**Returns <code>true</code> if the info property of the image contains no key/value pairs separated by an
    * equals signs.
    * @return <code>true</code> or <code>false</code>*/
   @SuppressWarnings("unused")
   public boolean isEmpty() {
      return map.isEmpty();
   }

   /**Returns the key set found in the info property of the image
    * @return the set*/
   @SuppressWarnings("unused")
   public Set<String> keySet() {
      return map.keySet();
   }

   @Override
   public String toString() {
      StringBuilder sb = new StringBuilder("InfoProperty{");
      sb.append(imp.getTitle());
      sb.append("}");
      sb.append("[");
      sb.append(map.size());
      sb.append("]");
      for (String key : map.keySet()) {
         sb.append("\n");
         sb.append("Info key: ");
         sb.append(key);
         sb.append(" value: ");
         sb.append(map.get(key));
      }
      return sb.toString();
   }
}
